/**
 * File: cdemu.c
 * Authors: Robert Penz <robert.penz@outertech.com>
 *          Mike Frysinger <vapier@gmail.com>
 * Copyright 2002-2004
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the Free Software           *
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             *
 *                                                                         *
 ***************************************************************************/

/*
 * Version 0.7 - 2004-12-8
 * works with current 2.6 kernels
 *
 * Version 0.6 - 2003-10-13
 * add linux-2.6.x support and break up code to try to
 * blur the differences between 2.4.x and 2.6.x
 *
 * Version 0.5 - 2003-07-02
 * rename to CDemu
 * rewrite of major parts to be 100% compatible to a cdrom
 * user space interface changed for new userspace program

 * Version 0.4 - 2003-04-08
 * removed sys_call_table use
 * fixed bug with devfsd
 * some extra error handling for the userspace program

 * Version 0.3 - 2003-01-11
 * added check if a "cd is loaded"
 * removed all warnings
 * code cleanup

 * Version 0.2 - 2003-01-12
 * support for proc fs
 * changed the vclib.c so it can be used by gui programs

 * Version 0.1 - 2003-01-10
 * first test release (data and raw mode work - but no checks, if you do 
 * something wrong you'll kill the kernel ;-)
 *
 */

#include "cdemu_kernel.h"
#include "cdemu.h"

extern struct cdrom_device_ops cdemu_dops;
extern struct block_device_operations cdemu_bdops;
extern void cdemu_request(request_queue_t*);
extern int cdemu_make_request(request_queue_t*, struct bio*);

int cdemu_major = MAJOR_NR;
int cdemu_count = 8;
struct cdemu_device *cdemu_dev;

static int __init cdemu_init(void)
{
/* check if the kernelparmeter is correct and register the block devices */
#define	MIN_CDEMUS	1		/* should be self explanatory ... */
#define	MAX_CDEMUS	255		/* 2.6 can have more than this but who would use it ? */

	int i;
	struct cdemu_device *vc;

	dprintk("cdemu_init: start\n");

	if ((cdemu_count < MIN_CDEMUS) || (cdemu_count > MAX_CDEMUS)) {
		printk_cdemu(KERN_WARNING, "invalid cdemu_count (must be between %i and %i), using default (%i)\n",
		             MIN_CDEMUS, MAX_CDEMUS, MIN_CDEMUS);
		cdemu_count = 8;
	}

	i = 
	register_blkdev(cdemu_major, DEVICE_NAME);
	if (i < 0) {
		printk_cdemu(KERN_ERR, "unable to register block device #%i for cdemu device\n", cdemu_major);
		return -EBUSY;
	} else if (i)
		cdemu_major = i;

	/* allocate cdemu_count entries in our struct */
	cdemu_dev = kmalloc(cdemu_count * sizeof(struct cdemu_device), GFP_KERNEL);
	if (!cdemu_dev) {
		printk_cdemu(KERN_ERR, "could not kmalloc enough memory for %i cdemu devices\n", cdemu_count);
		goto out_nomem;
	}

#if defined(CONFIG_DEVFS_FS)
	devfs_mk_dir("cdemu");
#endif

	for (i = 0; i < cdemu_count; ++i) {
		vc = &cdemu_dev[i];
		memset(vc, 0, sizeof(struct cdemu_device));
		vc->number = i;
		vc->loaded = 0;
		vc->media_changed = 0;
		vc->locked = 0;
		vc->pause = 0;
		vc->bdev = NULL;
		vc->info.ops = &cdemu_dops;
		vc->info.speed = 0; /* FIXME max_speed */
		vc->info.capacity = 1;
		vc->info.handle = vc;
		sprintf(vc->info.name, "%s%i", DEVICE_NAME, i);
		spin_lock_init(&vc->lock);
		dprintk("cdemu_init: making cdrom #%i [%p]: %s\n", vc->number, vc, vc->info.name);

		/* register with the uniform cdrom driver */
		vc->gd = alloc_disk(1);
		if (!vc->gd)
			goto out;
		vc->gd->major = cdemu_major;
		vc->gd->first_minor = i;
		sprintf(vc->gd->disk_name, "%s%i", DEVICE_NAME, i);
		vc->gd->fops = &cdemu_bdops;
		vc->gd->private_data = vc;
		vc->gd->flags = GENHD_FL_REMOVABLE | GENHD_FL_CD;
		set_capacity(vc->gd, 0);
		register_cdrom(&vc->info);
		vc->queue = blk_alloc_queue(GFP_KERNEL);
		blk_queue_make_request(vc->queue, cdemu_make_request);
		if (!vc->queue)
			goto out;
		vc->gd->queue = vc->queue;
		add_disk(vc->gd);
#ifdef CONFIG_DEVFS_FS
		devfs_mk_bdev(MKDEV(vc->gd->major, vc->gd->first_minor),
		              S_IFBLK|S_IRUSR|S_IWUSR,
		              "cdemu/%i", i);
#endif
	}

	if (!cdemu_proc_make()) {
		dprintk("cdemu_init: Error registering /proc entry!\n");
		goto out_freeall;
	}

	/* all done report our success */
	dprintk("cdemu_init: loaded (max %i devices)\n", cdemu_count);
	dprintk("cdemu_init: end\n");
	printk_cdemu(KERN_INFO, "ver " CDEMU_MODULE_VERSION " loaded.  Registered %i cdemus.\n", cdemu_count); 
	return 0;

out:
	for (i = 0; i < cdemu_count; ++i) {
		vc = &cdemu_dev[i];
		if (vc->queue)
			blk_put_queue(vc->queue);
		if (vc->gd) {
			put_disk(vc->gd);
			del_gendisk(vc->gd);
		}
	}
out_freeall:
	kfree(cdemu_dev);
out_nomem:
	if (unregister_blkdev(cdemu_major, DEVICE_NAME))
		printk_cdemu(KERN_ERR, "unable to unregister block device\n");

	printk_cdemu(KERN_ERR, "cdemu_init: ran out of memory\n");
	return -ENOMEM;
}

static void __exit cdemu_exit(void)
{
	int i;
	struct cdemu_device *vc;

	dprintk("cdemu_exit: start\n");

	/* get rid of the proc stuff */
	cdemu_proc_remove();

	for (i = 0; i < cdemu_count; ++i) {
		vc = &cdemu_dev[i];
		dprintk("cdemu_exit: unregistering cdrom #%i [%p]: %s\n", vc->number, vc, vc->info.name);
#ifdef CONFIG_DEVFS_FS
		devfs_remove("cdemu/%i", i);
#endif
		if (vc->bdev) {
			dprintk("cdemu_exit: punting bdev\n");
			invalidate_bdev(vc->bdev, 1);
			bdput(vc->bdev);
		}
		if (vc->gd) {
			dprintk("cdemu_exit: punting gendisk\n");
			del_gendisk(vc->gd);
			put_disk(vc->gd);
		}
		if (vc->queue) {
			dprintk("cdemu_exit: punting queue\n");
			blk_put_queue(vc->queue);
		}
		if (unregister_cdrom(&vc->info))
			printk(KERN_WARNING "cdemu: cannot unregister cdrom\n");
	}

#if defined(CONFIG_DEVFS_FS)
	devfs_remove("cdemu");
#endif

	if (unregister_blkdev(cdemu_major, DEVICE_NAME))
		printk_cdemu(KERN_ERR, "unable to unregister block device\n");

	kfree(cdemu_dev);

	dprintk("cdemu_exit: end\n");
	printk_cdemu(KERN_INFO, "ver " CDEMU_MODULE_VERSION " unloaded.  Unregistered %i cdemus.\n", cdemu_count); 
}

module_init(cdemu_init);
module_exit(cdemu_exit);

#if !defined(MODULE)
static int __init cdemu_count_setup(char *str)
{
	cdemu_count = simple_strtol(str, NULL, 0);
	return 1;
}

__setup("cdemu_count=", cdemu_count_setup);
#endif

MODULE_AUTHOR("Robert Penz <robert.penz@outertech.com>, Mike Frysinger <vapier@gmail.com>");
MODULE_DESCRIPTION("CD Emulator for bin/cue images ver " CDEMU_MODULE_VERSION);
MODULE_LICENSE("GPL");
module_param(cdemu_major, int, 0);
module_param(cdemu_count, int, 0);
MODULE_PARM_DESC(cdemu_count, "Maximum number of virtual cdroms (1-255)");
